/* Copyright 2002-2010 CS Communication & Systèmes
 * Licensed to CS Communication & Systèmes (CS) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * CS licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.orekit.attitudes;


import org.apache.commons.math.geometry.Rotation;
import org.apache.commons.math.geometry.Vector3D;
import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.orekit.Utils;
import org.orekit.bodies.GeodeticPoint;
import org.orekit.bodies.OneAxisEllipsoid;
import org.orekit.errors.OrekitException;
import org.orekit.frames.Frame;
import org.orekit.frames.FramesFactory;
import org.orekit.frames.Transform;
import org.orekit.orbits.CircularOrbit;
import org.orekit.orbits.KeplerianOrbit;
import org.orekit.propagation.Propagator;
import org.orekit.propagation.SpacecraftState;
import org.orekit.propagation.analytical.KeplerianPropagator;
import org.orekit.time.AbsoluteDate;
import org.orekit.time.DateComponents;
import org.orekit.time.TimeComponents;
import org.orekit.time.TimeScalesFactory;
import org.orekit.utils.Line;
import org.orekit.utils.PVCoordinates;


public class TargetPointingTest {

    // Computation date 
    private AbsoluteDate date;
    
    // Body mu 
    private double mu;

    // Reference frame = ITRF 2005C 
    private Frame frameITRF2005;
        
    // Transform from EME2000 to ITRF2005C 
    private Transform eme2000ToItrf;

    /** Test if both constructors are equivalent
     */
    @Test
    public void testConstructors() throws OrekitException {

        //  Satellite position
        // ********************
        CircularOrbit circ =
            new CircularOrbit(7178000.0, 0.5e-4, -0.5e-4, Math.toRadians(50.), Math.toRadians(270.),
                                   Math.toRadians(5.300), CircularOrbit.MEAN_LONGITUDE_ARGUMENT, 
                                   FramesFactory.getEME2000(), date, mu);
        
        //  Attitude laws
        // *************** 
        // Elliptic earth shape
        OneAxisEllipsoid earthShape = new OneAxisEllipsoid(6378136.460, 1 / 298.257222101, frameITRF2005);
                
        // Target definition as a geodetic point AND as a position/velocity vector
        GeodeticPoint geoTargetITRF2005C = new GeodeticPoint(Math.toRadians(43.36), Math.toRadians(1.26), 600.);
        Vector3D pTargetITRF2005C = earthShape.transform(geoTargetITRF2005C);
            
        // Attitude law definition from geodetic point target 
        TargetPointing geoTargetAttitudeLaw = new TargetPointing(geoTargetITRF2005C, earthShape);
        
        //  Attitude law definition from position/velocity target
        TargetPointing pvTargetAttitudeLaw = new TargetPointing(frameITRF2005, pTargetITRF2005C);
        
        // Check that both attitude are the same 
        // Get satellite rotation for target pointing law 
        Rotation rotPv = pvTargetAttitudeLaw.getAttitude(circ).getRotation();

        // Get satellite rotation for nadir pointing law
        Rotation rotGeo = geoTargetAttitudeLaw.getAttitude(circ).getRotation();

        // Rotations composition
        Rotation rotCompo = rotGeo.applyInverseTo(rotPv);
        double angle = rotCompo.getAngle();
        Assert.assertEquals(angle, 0.0, Utils.epsilonAngle);

    }

    /** Test if geodetic constructor works
     */
    @Test
    public void testGeodeticConstructor() throws OrekitException {

        //  Satellite position
        // ******************** 
        CircularOrbit circ =
            new CircularOrbit(7178000.0, 0.5e-4, -0.5e-4, Math.toRadians(50.), Math.toRadians(270.),
                                   Math.toRadians(5.300), CircularOrbit.MEAN_LONGITUDE_ARGUMENT, 
                                   FramesFactory.getEME2000(), date, mu);
        
        //  Attitude law
        // ************** 
        
        // Elliptic earth shape 
        OneAxisEllipsoid earthShape = new OneAxisEllipsoid(6378136.460, 1 / 298.257222101, frameITRF2005);
                
        // Target definition as a geodetic point 
        GeodeticPoint geoTargetITRF2005 = new GeodeticPoint(Math.toRadians(43.36), Math.toRadians(1.26), 600.);
            
        //  Attitude law definition 
        TargetPointing geoTargetAttitudeLaw = new TargetPointing(geoTargetITRF2005, earthShape);
        
        // Check that observed ground point is the same as defined target 
        Vector3D pObservedEME2000 = geoTargetAttitudeLaw.getTargetPoint(circ, FramesFactory.getEME2000());
        GeodeticPoint geoObserved = earthShape.transform(pObservedEME2000, FramesFactory.getEME2000(), date);

        Assert.assertEquals(geoObserved.getLongitude(), geoTargetITRF2005.getLongitude(), Utils.epsilonAngle);
        Assert.assertEquals(geoObserved.getLatitude(), geoTargetITRF2005.getLatitude(), Utils.epsilonAngle);
        Assert.assertEquals(geoObserved.getAltitude(), geoTargetITRF2005.getAltitude(), 1.e-8);

    }

    /** Test with nadir target : Check that when the target is the same as nadir target at date,
     * satellite attitude is the same as nadir attitude at the same date, but different at a different date.
     */
    @Test
    public void testNadirTarget() throws OrekitException {

        // Elliptic earth shape
        OneAxisEllipsoid earthShape = new OneAxisEllipsoid(6378136.460, 1 / 298.257222101, frameITRF2005);
                
        // Satellite on any position 
        CircularOrbit circOrbit =
            new CircularOrbit(7178000.0, 1.e-5, 0., Math.toRadians(50.), 0.,
                                   Math.toRadians(90.), CircularOrbit.TRUE_LONGITUDE_ARGUMENT, 
                                   FramesFactory.getEME2000(), date, mu);

        //  Target attitude law with target under satellite nadir 
        // ******************************************************* 
        // Definition of nadir target 
        // Create nadir pointing attitude law 
        NadirPointing nadirAttitudeLaw = new NadirPointing(earthShape);
        
        // Check nadir target 
        Vector3D pNadirTarget  = nadirAttitudeLaw.getTargetPoint(circOrbit,  frameITRF2005);
        GeodeticPoint geoNadirTarget = earthShape.transform(pNadirTarget, frameITRF2005, date);
        
        // Create target attitude law 
        TargetPointing targetAttitudeLaw = new TargetPointing(geoNadirTarget, earthShape);

        //  1/ Test that attitudes are the same at date
        // *********************************************
        // i.e the composition of inverse earth pointing rotation
        // with nadir pointing rotation shall be identity. 
        
        // Get satellite rotation from target pointing law at date
        Rotation rotTarget = targetAttitudeLaw.getAttitude(circOrbit).getRotation();

        // Get satellite rotation from nadir pointing law at date
        Rotation rotNadir = nadirAttitudeLaw.getAttitude(circOrbit).getRotation();

        // Compose attitude rotations
        Rotation rotCompo = rotTarget.applyInverseTo(rotNadir);
        double angle = rotCompo.getAngle();
        Assert.assertEquals(angle, 0.0, Utils.epsilonAngle);

        
        //  2/ Test that attitudes are different at a different date
        // **********************************************************

        // Extrapolation one minute later
        KeplerianPropagator extrapolator = new KeplerianPropagator(circOrbit);
        double delta_t = 60.0; // extrapolation duration in seconds
        AbsoluteDate extrapDate = date.shiftedBy(delta_t);
        SpacecraftState extrapOrbit = extrapolator.propagate(extrapDate);
        
        // Get satellite rotation from target pointing law at date + 1min
        Rotation extrapRotTarget = targetAttitudeLaw.getAttitude(extrapOrbit.getOrbit()).getRotation();
        
        // Get satellite rotation from nadir pointing law at date
        Rotation extrapRotNadir = nadirAttitudeLaw.getAttitude(extrapOrbit.getOrbit()).getRotation();

        // Compose attitude rotations
        Rotation extrapRotCompo = extrapRotTarget.applyInverseTo(extrapRotNadir);
        double extrapAngle = extrapRotCompo.getAngle();
        Assert.assertEquals(extrapAngle, Math.toRadians(24.684793905118823), Utils.epsilonAngle);
        
    }
       
    /** Test if defined target belongs to the direction pointed by the satellite
     */
    @Test
    public void testTargetInPointingDirection() throws OrekitException {

        // Create computation date 
        AbsoluteDate date = new AbsoluteDate(new DateComponents(2008, 04, 07),
                                             TimeComponents.H00,
                                             TimeScalesFactory.getUTC());
        
        // Reference frame = ITRF 2005
        Frame frameITRF2005 = FramesFactory.getITRF2005(true);

        // Elliptic earth shape 
        OneAxisEllipsoid earthShape = new OneAxisEllipsoid(6378136.460, 1 / 298.257222101, frameITRF2005);
                
        // Create target pointing attitude law 
        GeodeticPoint geoTarget = new GeodeticPoint(Math.toRadians(43.36), Math.toRadians(1.26), 600.);
        TargetPointing targetAttitudeLaw = new TargetPointing(geoTarget, earthShape);
        
        //  Satellite position
        // ********************
        // Create satellite position as circular parameters
        CircularOrbit circ =
            new CircularOrbit(7178000.0, 0.5e-4, -0.5e-4, Math.toRadians(50.), Math.toRadians(270.),
                                   Math.toRadians(5.300), CircularOrbit.MEAN_LONGITUDE_ARGUMENT, 
                                   FramesFactory.getEME2000(), date, mu);
        
        // Transform satellite position to position/velocity parameters in EME2000 frame
        PVCoordinates pvSatEME2000 = circ.getPVCoordinates();
        
        //  Pointing direction
        // ********************
        // Get satellite attitude rotation, i.e rotation from EME2000 frame to satellite frame
        Rotation rotSatEME2000 = targetAttitudeLaw.getAttitude(circ).getRotation();
        
        // Transform Z axis from satellite frame to EME2000 
        Vector3D zSatEME2000 = rotSatEME2000.applyInverseTo(Vector3D.PLUS_K);
        
        // Line containing satellite point and following pointing direction
        Line pointingLine = new Line(eme2000ToItrf.transformPosition(pvSatEME2000.getPosition()), eme2000ToItrf.transformVector(zSatEME2000));
        
        // Check that the line contains earth center
        double distance = pointingLine.distance(earthShape.transform(geoTarget));
        
        Assert.assertTrue(distance < 1.e-7);
    }

    /** Test the difference between pointing over two longitudes separated by 5°
     */
    @Test
    public void testSlewedTarget() throws OrekitException {

        // Spheric earth shape 
        OneAxisEllipsoid earthShape = new OneAxisEllipsoid(6378136.460, 0., frameITRF2005);
        
        //  Satellite position
        // ********************
        // Create satellite position as circular parameters
        CircularOrbit circ =
            new CircularOrbit(42164000.0, 0.5e-8, -0.5e-8, 0., 0.,
                                   Math.toRadians(5.300), CircularOrbit.MEAN_LONGITUDE_ARGUMENT, 
                                   FramesFactory.getEME2000(), date, mu);
        
        // Create nadir pointing attitude law 
        // ********************************** 
        NadirPointing nadirAttitudeLaw = new NadirPointing(earthShape);
        
        // Get observed ground point from nadir pointing law
        Vector3D pNadirObservedEME2000 = nadirAttitudeLaw.getTargetPoint(circ, FramesFactory.getEME2000());
        Vector3D pNadirObservedITRF2005 = eme2000ToItrf.transformPosition(pNadirObservedEME2000);
        
        GeodeticPoint geoNadirObserved = earthShape.transform(pNadirObservedITRF2005, frameITRF2005, date);

        // Create target pointing attitude law with target equal to nadir target 
        // ********************************************************************* 
        TargetPointing targetLawRef = new TargetPointing(frameITRF2005, pNadirObservedITRF2005);
        
        // Get attitude rotation in EME2000
        Rotation rotSatRefEME2000 = targetLawRef.getAttitude(circ).getRotation();
      
        // Create target pointing attitude law with target 5° from nadir target 
        // ******************************************************************** 
        GeodeticPoint geoTarget = new GeodeticPoint(geoNadirObserved.getLatitude(),
                                                    geoNadirObserved.getLongitude() - Math.toRadians(5), geoNadirObserved.getAltitude());
        Vector3D pTargetITRF2005C = earthShape.transform(geoTarget);
        TargetPointing targetLaw = new TargetPointing(frameITRF2005, pTargetITRF2005C);
        
        // Get attitude rotation 
        Rotation rotSatEME2000 = targetLaw.getAttitude(circ).getRotation();
        
        // Compute difference between both attitude laws 
        // *********************************************
        // Difference between attitudes
        //  expected
        double tanDeltaExpected = (6378136.460/(42164000.0-6378136.460))*Math.tan(Math.toRadians(5));
        double deltaExpected = Math.atan(tanDeltaExpected);
         
        //  real
        double deltaReal = rotSatEME2000.applyInverseTo(rotSatRefEME2000).getAngle();
        
        Assert.assertEquals(deltaReal, deltaExpected, 1.e-4);
        
    } 

    @Test
    public void testSpin() throws OrekitException {

        Frame frameITRF2005 = FramesFactory.getITRF2005(true);

        // Elliptic earth shape 
        OneAxisEllipsoid earthShape = new OneAxisEllipsoid(6378136.460, 1 / 298.257222101, frameITRF2005);
                
        // Create target pointing attitude law 
        GeodeticPoint geoTarget = new GeodeticPoint(Math.toRadians(43.36), Math.toRadians(1.26), 600.);
        AttitudeLaw law = new TargetPointing(geoTarget, earthShape);

        KeplerianOrbit orbit =
            new KeplerianOrbit(7178000.0, 1.e-4, Math.toRadians(50.),
                              Math.toRadians(10.), Math.toRadians(20.),
                              Math.toRadians(30.), KeplerianOrbit.MEAN_ANOMALY, 
                              FramesFactory.getEME2000(), date, 3.986004415e14);

        Propagator propagator = new KeplerianPropagator(orbit, law);

        double h = 0.01;
        SpacecraftState sMinus = propagator.propagate(date.shiftedBy(-h));
        SpacecraftState s0     = propagator.propagate(date);
        SpacecraftState sPlus  = propagator.propagate(date.shiftedBy(h));

        Vector3D spin0 = s0.getAttitude().getSpin();
        Vector3D reference = Attitude.estimateSpin(sMinus.getAttitude().getRotation(),
                                                   sPlus.getAttitude().getRotation(),
                                                   2 * h);
        Assert.assertEquals(0.0, spin0.subtract(reference).getNorm(), 1.1e-10);

    }

    @Before
    public void setUp() {
        try {

            Utils.setDataRoot("regular-data");

            // Computation date
            date = new AbsoluteDate(new DateComponents(2008, 04, 07),
                                    TimeComponents.H00,
                                    TimeScalesFactory.getUTC());

            // Body mu
            mu = 3.9860047e14;
            
            // Reference frame = ITRF 2005
            frameITRF2005 = FramesFactory.getITRF2005(true);

            // Transform from EME2000 to ITRF2005
            eme2000ToItrf = FramesFactory.getEME2000().getTransformTo(frameITRF2005, date);

        } catch (OrekitException oe) {
            Assert.fail(oe.getMessage());
        }

    }

    @After
    public void tearDown() {
        date = null;
        frameITRF2005 = null;
        eme2000ToItrf = null;
    }

}

